"""maleware detection exceptions module."""


class MalwareDetectionError(Exception):
    """Base exception for all malware detection exceptions."""


class MalwareDetectionInvalidAuthenticationError(MalwareDetectionError):
    """Exception raised when the authentication is invalid."""


class MalwareDetectionInvalidBackendError(MalwareDetectionError):
    """Exception raised when the backend is invalid."""
